# -*- coding: utf-8 -*-

import pytest

from marshmallow import Schema, fields, class_registry
from marshmallow.exceptions import RegistryError


def test_serializer_has_class_registry():
    class MySchema(Schema):
        pass

    class MySubSchema(Schema):
        pass

    assert 'MySchema' in class_registry._registry
    assert 'MySubSchema' in class_registry._registry


class A:

    def __init__(self, _id, b=None):
        self.id = _id
        self.b = b

class B:

    def __init__(self, _id, a=None):
        self.id = _id
        self.a = a


class C:

    def __init__(self, _id, bs=None):
        self.id = _id
        self.bs = bs or []

class ASchema(Schema):
    id = fields.Integer()
    b = fields.Nested('BSchema', exclude=('a', ))

class BSchema(Schema):
    id = fields.Integer()
    a = fields.Nested('ASchema')

class CSchema(Schema):
    id = fields.Integer()
    bs = fields.Nested('BSchema', many=True)

def test_two_way_nesting():
    a_obj = A(1)
    b_obj = B(2, a=a_obj)
    a_obj.b = b_obj

    a_serialized = ASchema().dump(a_obj)
    b_serialized = BSchema().dump(b_obj)
    assert a_serialized['b']['id'] == b_obj.id
    assert b_serialized['a']['id'] == a_obj.id

def test_nesting_with_class_name_many():
    c_obj = C(1, bs=[B(2), B(3), B(4)])

    c_serialized = CSchema().dump(c_obj)

    assert len(c_serialized['bs']) == len(c_obj.bs)
    assert c_serialized['bs'][0]['id'] == c_obj.bs[0].id

def test_invalid_class_name_in_nested_field_raises_error(user):

    class MySchema(Schema):
        nf = fields.Nested('notfound')
    sch = MySchema()
    with pytest.raises(RegistryError) as excinfo:
        sch.dump({'nf': None})
    assert 'Class with name {0!r} was not found'.format('notfound') in str(excinfo)

class FooSerializer(Schema):
    _id = fields.Integer()


def test_multiple_classes_with_same_name_raises_error():
    # Import a class with the same name
    from .foo_serializer import FooSerializer as FooSerializer1  # noqa

    class MySchema(Schema):
        foo = fields.Nested('FooSerializer')

    # Using a nested field with the class name fails because there are
    # two defined classes with the same name
    sch = MySchema()
    with pytest.raises(RegistryError) as excinfo:
        sch.dump({'foo': {'_id': 1}})
    msg = 'Multiple classes with name {0!r} were found.'\
            .format('FooSerializer')
    assert msg in str(excinfo)

def test_multiple_classes_with_all():
    # Import a class with the same name
    from .foo_serializer import FooSerializer as FooSerializer1  # noqa

    classes = class_registry.get_class('FooSerializer', all=True)
    assert len(classes) == 2


def test_can_use_full_module_path_to_class():
    from .foo_serializer import FooSerializer as FooSerializer1  # noqa
    # Using full paths is ok

    class Schema1(Schema):
        foo = fields.Nested('tests.foo_serializer.FooSerializer')

    sch = Schema1()

    # Note: The arguments here don't matter. What matters is that no
    # error is raised
    assert sch.dump({'foo': {'_id': 42}})

    class Schema2(Schema):
        foo = fields.Nested('tests.test_registry.FooSerializer')
    sch = Schema2()
    assert sch.dump({'foo': {'_id': 42}})



# -*- coding: utf-8 -*-
import datetime as dt
from collections import namedtuple
from functools import partial

import pytest

from marshmallow import utils
from tests.base import (
    assert_datetime_equal,
    central,
    assert_time_equal,
    assert_date_equal,
)


def test_to_marshallable_type():
    class Foo(object):
        CLASS_VAR = 'bar'

        def __init__(self):
            self.attribute = 'baz'

        @property
        def prop(self):
            return 42

    obj = Foo()
    u_dict = utils.to_marshallable_type(obj)
    assert u_dict['CLASS_VAR'] == Foo.CLASS_VAR
    assert u_dict['attribute'] == obj.attribute
    assert u_dict['prop'] == obj.prop

def test_to_marshallable_type_none():
    assert utils.to_marshallable_type(None) is None


PointNT = namedtuple('Point', ['x', 'y'])

def test_to_marshallable_type_with_namedtuple():
    p = PointNT(24, 42)
    result = utils.to_marshallable_type(p)
    assert result['x'] == p.x
    assert result['y'] == p.y

class PointClass(object):
    def __init__(self, x, y):
        self.x = x
        self.y = y

@pytest.mark.parametrize('obj', [
    PointNT(24, 42),
    PointClass(24, 42),
    {'x': 24, 'y': 42}
])
def test_get_value_from_object(obj):
    result = utils.get_value(obj, 'x')
    assert result == 24
    result2 = utils.get_value(obj, 'y')
    assert result2 == 42

def test_get_value_from_namedtuple_with_default():
    p = PointNT(x=42, y=None)
    # Default is only returned if key is not found
    assert utils.get_value(p, 'z', default=123) == 123
    # since 'y' is an attribute, None is returned instead of the default
    assert utils.get_value(p, 'y', default=123) is None

class Triangle(object):
    def __init__(self, p1, p2, p3):
        self.p1 = p1
        self.p2 = p2
        self.p3 = p3
        self.points = [p1, p2, p3]

def test_get_value_for_nested_object():
    tri = Triangle(p1=PointClass(1, 2), p2=PointNT(3, 4), p3={'x': 5, 'y': 6})
    assert utils.get_value(tri, 'p1.x') == 1
    assert utils.get_value(tri, 'p2.x') == 3
    assert utils.get_value(tri, 'p3.x') == 5

# regression test for https://github.com/marshmallow-code/marshmallow/issues/62
def test_get_value_from_dict():
    d = dict(items=['foo', 'bar'], keys=['baz', 'quux'])
    assert utils.get_value(d, 'items') == ['foo', 'bar']
    assert utils.get_value(d, 'keys') == ['baz', 'quux']

def test_get_value():
    lst = [1, 2, 3]
    assert utils.get_value(lst, 1) == 2

    class MyInt(int):
        pass

    assert utils.get_value(lst, MyInt(1)) == 2


def test_set_value():
    d = {}
    utils.set_value(d, 'foo', 42)
    assert d == {'foo': 42}

    d = {}
    utils.set_value(d, 'foo.bar', 42)
    assert d == {'foo': {'bar': 42}}

    d = {'foo': {}}
    utils.set_value(d, 'foo.bar', 42)
    assert d == {'foo': {'bar': 42}}

    d = {'foo': 42}
    with pytest.raises(ValueError):
        utils.set_value(d, 'foo.bar', 42)


def test_is_keyed_tuple():
    Point = namedtuple('Point', ['x', 'y'])
    p = Point(24, 42)
    assert utils.is_keyed_tuple(p) is True
    t = (24, 42)
    assert utils.is_keyed_tuple(t) is False
    d = {'x': 42, 'y': 24}
    assert utils.is_keyed_tuple(d) is False
    s = 'xy'
    assert utils.is_keyed_tuple(s) is False
    lst = [24, 42]
    assert utils.is_keyed_tuple(lst) is False

def test_to_marshallable_type_list():
    assert utils.to_marshallable_type(['foo', 'bar']) == ['foo', 'bar']

def test_to_marshallable_type_generator():
    gen = (e for e in ['foo', 'bar'])
    assert utils.to_marshallable_type(gen) == ['foo', 'bar']

def test_marshallable():
    class ObjContainer(object):
        contained = {"foo": 1}

        def __marshallable__(self):
            return self.contained

    obj = ObjContainer()
    assert utils.to_marshallable_type(obj) == {"foo": 1}

def test_is_collection():
    assert utils.is_collection([1, 'foo', {}]) is True
    assert utils.is_collection(('foo', 2.3)) is True
    assert utils.is_collection({'foo': 'bar'}) is False

def test_rfcformat_gmt_naive():
    d = dt.datetime(2013, 11, 10, 1, 23, 45)
    assert utils.rfcformat(d) == "Sun, 10 Nov 2013 01:23:45 -0000"

def test_rfcformat_central():
    d = central.localize(dt.datetime(2013, 11, 10, 1, 23, 45), is_dst=False)
    assert utils.rfcformat(d) == 'Sun, 10 Nov 2013 07:23:45 -0000'

def test_rfcformat_central_localized():
    d = central.localize(dt.datetime(2013, 11, 10, 8, 23, 45), is_dst=False)
    assert utils.rfcformat(d, localtime=True) == "Sun, 10 Nov 2013 08:23:45 -0600"

def test_isoformat():
    d = dt.datetime(2013, 11, 10, 1, 23, 45)
    assert utils.isoformat(d) == '2013-11-10T01:23:45+00:00'

def test_isoformat_tzaware():
    d = central.localize(dt.datetime(2013, 11, 10, 1, 23, 45), is_dst=False)
    assert utils.isoformat(d) == "2013-11-10T07:23:45+00:00"

def test_isoformat_localtime():
    d = central.localize(dt.datetime(2013, 11, 10, 1, 23, 45), is_dst=False)
    assert utils.isoformat(d, localtime=True) == "2013-11-10T01:23:45-06:00"

def test_from_datestring():
    d = dt.datetime.now()
    rfc = utils.rfcformat(d)
    iso = d.isoformat()
    assert_date_equal(utils.from_datestring(rfc), d)
    assert_date_equal(utils.from_datestring(iso), d)

@pytest.mark.parametrize('use_dateutil', [True, False])
def test_from_rfc(use_dateutil):
    d = dt.datetime.now()
    rfc = utils.rfcformat(d)
    result = utils.from_rfc(rfc, use_dateutil=use_dateutil)
    assert type(result) == dt.datetime
    assert_datetime_equal(result, d)

@pytest.mark.parametrize('use_dateutil', [True, False])
def test_from_iso_datetime(use_dateutil):
    d = dt.datetime.now()
    formatted = d.isoformat()
    result = utils.from_iso_datetime(formatted, use_dateutil=use_dateutil)
    assert type(result) == dt.datetime
    assert_datetime_equal(result, d)

def test_from_iso_with_tz():
    d = central.localize(dt.datetime.now())
    formatted = d.isoformat()
    result = utils.from_iso_datetime(formatted)
    assert_datetime_equal(result, d)
    if utils.dateutil_available:
        # Note a naive datetime
        assert result.tzinfo is not None

# Test with and without dateutil
@pytest.mark.parametrize('use_dateutil', [True, False])
def test_from_iso_time_with_microseconds(use_dateutil):
    t = dt.time(1, 23, 45, 6789)
    formatted = t.isoformat()
    result = utils.from_iso_time(formatted, use_dateutil=use_dateutil)
    assert type(result) == dt.time
    assert_time_equal(result, t, microseconds=True)

@pytest.mark.parametrize('use_dateutil', [True, False])
def test_from_iso_time_without_microseconds(use_dateutil):
    t = dt.time(1, 23, 45)
    formatted = t.isoformat()
    result = utils.from_iso_time(formatted, use_dateutil=use_dateutil)
    assert type(result) == dt.time
    assert_time_equal(result, t, microseconds=True)

@pytest.mark.parametrize('use_dateutil', [True, False])
def test_from_iso_date(use_dateutil):
    d = dt.date(2014, 8, 21)
    iso_date = d.isoformat()
    result = utils.from_iso_date(iso_date, use_dateutil=use_dateutil)
    assert type(result) == dt.date
    assert_date_equal(result, d)

def test_get_func_args():
    def f1(foo, bar):
        pass

    f2 = partial(f1, 'baz')

    class F3(object):
        def __call__(self, foo, bar):
            pass
    f3 = F3()

    for func in [f1, f2, f3]:
        assert utils.get_func_args(func) == ['foo', 'bar']



# -*- coding: utf-8 -*-
"""Tests for marshmallow.validate"""
from __future__ import unicode_literals

import re
import pytest

from marshmallow.compat import PY2
from marshmallow import validate, ValidationError

@pytest.mark.parametrize('valid_url', [
    'http://example.org',
    'https://example.org',
    'ftp://example.org',
    'ftps://example.org',
    'http://example.co.jp',
    'http://www.example.com/a%C2%B1b',
    'http://www.example.com/~username/',
    'http://info.example.com/?fred',
    'http://xn--mgbh0fb.xn--kgbechtv/',
    'http://example.com/blue/red%3Fand+green',
    'http://www.example.com/?array%5Bkey%5D=value',
    'http://xn--rsum-bpad.example.org/',
    'http://123.45.67.8/',
    'http://123.45.67.8:8329/',
    'http://[2001:db8::ff00:42]:8329',
    'http://[2001::1]:8329',
    'http://www.example.com:8000/foo',
])
def test_url_absolute_valid(valid_url):
    validator = validate.URL(relative=False)
    assert validator(valid_url) == valid_url

@pytest.mark.parametrize('invalid_url', [
    'http:///example.com/',
    'https:///example.com/',
    'https://example.org\\',
    'ftp:///example.com/',
    'ftps:///example.com/',
    'http//example.org',
    'http:///',
    'http:/example.org',
    'foo://example.org',
    '../icons/logo.gif',
    'http://2001:db8::ff00:42:8329',
    'http://[192.168.1.1]:8329',
    'abc',
    '..',
    '/',
    ' ',
    '',
    None,
])
def test_url_absolute_invalid(invalid_url):
    validator = validate.URL(relative=False)
    with pytest.raises(ValidationError):
        validator(invalid_url)

@pytest.mark.parametrize('valid_url', [
    'http://example.org',
    'http://123.45.67.8/',
    'http://example.com/foo/bar/../baz',
    'https://example.com/../icons/logo.gif',
    'http://example.com/./icons/logo.gif',
    'ftp://example.com/../../../../g',
    'http://example.com/g?y/./x',
])
def test_url_relative_valid(valid_url):
    validator = validate.URL(relative=True)
    assert validator(valid_url) == valid_url

@pytest.mark.parametrize('invalid_url', [
    'http//example.org',
    'suppliers.html',
    '../icons/logo.gif',
    '\icons/logo.gif',
    '../.../g',
    '...',
    '\\',
    ' ',
    '',
    None,
])
def test_url_relative_invalid(invalid_url):
    validator = validate.URL(relative=True)
    with pytest.raises(ValidationError):
        validator(invalid_url)

@pytest.mark.parametrize('valid_url', [
    'http://example.org',
    'http://123.45.67.8/',
    'http://example',
    'http://example.',
    'http://example:80',
    'http://user.name:pass.word@example',
    'http://example/foo/bar',
])
def test_url_dont_require_tld_valid(valid_url):
    validator = validate.URL(require_tld=False)
    assert validator(valid_url) == valid_url

@pytest.mark.parametrize('invalid_url', [
    'http//example',
    'http://.example.org',
    'http:///foo/bar',
    'http:// /foo/bar',
    '',
    None,
])
def test_url_dont_require_tld_invalid(invalid_url):
    validator = validate.URL(require_tld=False)
    with pytest.raises(ValidationError):
        validator(invalid_url)

def test_url_custom_scheme():
    validator = validate.URL()
    # By default, ws not allowed
    url = 'ws://test.test'
    with pytest.raises(ValidationError):
        validator(url)

    validator = validate.URL(schemes=set(['http', 'https', 'ws']))
    assert validator(url) == url

def test_url_relative_and_custom_schemes():
    validator = validate.URL(relative=True)
    # By default, ws not allowed
    url = 'ws://test.test'
    with pytest.raises(ValidationError):
        validator(url)

    validator = validate.URL(relative=True, schemes=set(['http', 'https', 'ws']))
    assert validator(url) == url

def test_url_custom_message():
    validator = validate.URL(error="{input} ain't an URL")
    with pytest.raises(ValidationError) as excinfo:
        validator('invalid')
    assert "invalid ain't an URL" in str(excinfo)

def test_url_repr():
    assert (
        repr(validate.URL(relative=False, error=None)) ==
        '<URL(relative=False, error={0!r})>'
        .format('Not a valid URL.')
    )
    assert (
        repr(validate.URL(relative=True, error='foo')) ==
        '<URL(relative=True, error={0!r})>'
        .format('foo')
    )


@pytest.mark.parametrize('valid_email', [
    'niceandsimple@example.com',
    'NiCeAnDsImPlE@eXaMpLe.CoM',
    'very.common@example.com',
    'a.little.lengthy.but.fine@a.iana-servers.net',
    'disposable.style.email.with+symbol@example.com',
    '"very.unusual.@.unusual.com"@example.com',
    "!#$%&'*+-/=?^_`{}|~@example.org",
    'niceandsimple@[64.233.160.0]',
    'niceandsimple@localhost',
    u'josé@blah.com',
    u'δοκ.ιμή@παράδειγμα.δοκιμή',
])
def test_email_valid(valid_email):
    validator = validate.Email()
    assert validator(valid_email) == valid_email

@pytest.mark.parametrize('invalid_email', [
    'a"b(c)d,e:f;g<h>i[j\\k]l@example.com',
    'just"not"right@example.com',
    'this is"not\allowed@example.com',
    'this\\ still\\"not\\\\allowed@example.com',
    '"much.more unusual"@example.com',
    '"very.(),:;<>[]\".VERY.\"very@\\ \"very\".unusual"@strange.example.com',
    '" "@example.org',
    'user@example',
    '@nouser.com',
    'example.com',
    'user',
    '',
    None,
])
def test_email_invalid(invalid_email):
    validator = validate.Email()
    with pytest.raises(ValidationError):
        validator(invalid_email)

def test_email_custom_message():
    validator = validate.Email(error='{input} is not an email addy.')
    with pytest.raises(ValidationError) as excinfo:
        validator('invalid')
    assert 'invalid is not an email addy.' in str(excinfo)

def test_email_repr():
    assert (
        repr(validate.Email(error=None)) ==
        '<Email(error={0!r})>'
        .format('Not a valid email address.')
    )
    assert (
        repr(validate.Email(error='foo')) ==
        '<Email(error={0!r})>'
        .format('foo')
    )


def test_range_min():
    assert validate.Range(1, 2)(1) == 1
    assert validate.Range(0)(1) == 1
    assert validate.Range()(1) == 1
    assert validate.Range(1, 1)(1) == 1

    with pytest.raises(ValidationError):
        validate.Range(2, 3)(1)
    with pytest.raises(ValidationError):
        validate.Range(2)(1)

def test_range_max():
    assert validate.Range(1, 2)(2) == 2
    assert validate.Range(None, 2)(2) == 2
    assert validate.Range()(2) == 2
    assert validate.Range(2, 2)(2) == 2

    with pytest.raises(ValidationError):
        validate.Range(0, 1)(2)
    with pytest.raises(ValidationError):
        validate.Range(None, 1)(2)

def test_range_custom_message():
    v = validate.Range(2, 3, error='{input} is not between {min} and {max}')
    with pytest.raises(ValidationError) as excinfo:
        v(1)
    assert '1 is not between 2 and 3' in str(excinfo)

    v = validate.Range(2, None, error='{input} is less than {min}')
    with pytest.raises(ValidationError) as excinfo:
        v(1)
    assert '1 is less than 2' in str(excinfo)

    v = validate.Range(None, 3, error='{input} is greater than {max}')
    with pytest.raises(ValidationError) as excinfo:
        v(4)
    assert '4 is greater than 3' in str(excinfo)

def test_range_repr():
    assert (
        repr(validate.Range(min=None, max=None, error=None)) ==
        '<Range(min=None, max=None, error=None)>'
    )
    assert (
        repr(validate.Range(min=1, max=3, error='foo')) ==
        '<Range(min=1, max=3, error={0!r})>'
        .format('foo')
    )


def test_length_min():
    assert validate.Length(3, 5)('foo') == 'foo'
    assert validate.Length(3, 5)([1, 2, 3]) == [1, 2, 3]
    assert validate.Length(0)('a') == 'a'
    assert validate.Length(0)([1]) == [1]
    assert validate.Length()('') == ''
    assert validate.Length()([]) == []
    assert validate.Length(1, 1)('a') == 'a'
    assert validate.Length(1, 1)([1]) == [1]

    with pytest.raises(ValidationError):
        validate.Length(4, 5)('foo')
    with pytest.raises(ValidationError):
        validate.Length(4, 5)([1, 2, 3])
    with pytest.raises(ValidationError):
        validate.Length(5)('foo')
    with pytest.raises(ValidationError):
        validate.Length(5)([1, 2, 3])

def test_length_max():
    assert validate.Length(1, 3)('foo') == 'foo'
    assert validate.Length(1, 3)([1, 2, 3]) == [1, 2, 3]
    assert validate.Length(None, 1)('a') == 'a'
    assert validate.Length(None, 1)([1]) == [1]
    assert validate.Length()('') == ''
    assert validate.Length()([]) == []
    assert validate.Length(2, 2)('ab') == 'ab'
    assert validate.Length(2, 2)([1, 2]) == [1, 2]

    with pytest.raises(ValidationError):
        validate.Length(1, 2)('foo')
    with pytest.raises(ValidationError):
        validate.Length(1, 2)([1, 2, 3])
    with pytest.raises(ValidationError):
        validate.Length(None, 2)('foo')
    with pytest.raises(ValidationError):
        validate.Length(None, 2)([1, 2, 3])

def test_length_equal():
    assert validate.Length(equal=3)('foo') == 'foo'
    assert validate.Length(equal=3)([1, 2, 3]) == [1, 2, 3]
    assert validate.Length(equal=None)('') == ''
    assert validate.Length(equal=None)([]) == []

    with pytest.raises(ValidationError):
        validate.Length(equal=2)('foo')
    with pytest.raises(ValidationError):
        validate.Length(equal=2)([1, 2, 3])

    with pytest.raises(ValueError):
        validate.Length(1, None, equal=3)('foo')
    with pytest.raises(ValueError):
        validate.Length(None, 5, equal=3)('foo')
    with pytest.raises(ValueError):
        validate.Length(1, 5, equal=3)('foo')

def test_length_custom_message():
    v = validate.Length(5, 6, error='{input} is not between {min} and {max}')
    with pytest.raises(ValidationError) as excinfo:
        v('foo')
    assert 'foo is not between 5 and 6' in str(excinfo)

    v = validate.Length(5, None, error='{input} is shorter than {min}')
    with pytest.raises(ValidationError) as excinfo:
        v('foo')
    assert 'foo is shorter than 5' in str(excinfo)

    v = validate.Length(None, 2, error='{input} is longer than {max}')
    with pytest.raises(ValidationError) as excinfo:
        v('foo')
    assert 'foo is longer than 2' in str(excinfo)

    v = validate.Length(None, None, equal=4, error='{input} does not have {equal}')
    with pytest.raises(ValidationError) as excinfo:
        v('foo')
    assert 'foo does not have 4' in str(excinfo)

def test_length_repr():
    assert (
        repr(validate.Length(min=None, max=None, error=None, equal=None)) ==
        '<Length(min=None, max=None, equal=None, error=None)>'
    )
    assert (
        repr(validate.Length(min=1, max=3, error='foo', equal=None)) ==
        '<Length(min=1, max=3, equal=None, error={0!r})>'
        .format('foo')
    )
    assert (
        repr(validate.Length(min=None, max=None, error='foo', equal=5)) ==
        '<Length(min=None, max=None, equal=5, error={0!r})>'
        .format('foo')
    )


def test_equal():
    assert validate.Equal('a')('a') == 'a'
    assert validate.Equal(1)(1) == 1
    assert validate.Equal([1])([1]) == [1]

    with pytest.raises(ValidationError):
        validate.Equal('b')('a')
    with pytest.raises(ValidationError):
        validate.Equal(2)(1)
    with pytest.raises(ValidationError):
        validate.Equal([2])([1])

def test_equal_custom_message():
    v = validate.Equal('a', error='{input} is not equal to {other}.')
    with pytest.raises(ValidationError) as excinfo:
        v('b')
    assert 'b is not equal to a.' in str(excinfo)

def test_equal_repr():
    assert (
        repr(validate.Equal(comparable=123, error=None)) ==
        '<Equal(comparable=123, error={0!r})>'
        .format('Must be equal to {other}.')
    )
    assert (
        repr(validate.Equal(comparable=123, error='foo')) ==
        '<Equal(comparable=123, error={0!r})>'
        .format('foo')
    )


def test_regexp_str():
    assert validate.Regexp(r'a')('a') == 'a'
    assert validate.Regexp(r'\w')('_') == '_'
    assert validate.Regexp(r'\s')(' ') == ' '
    assert validate.Regexp(r'1')('1') == '1'
    assert validate.Regexp(r'[0-9]+')('1') == '1'
    assert validate.Regexp(r'a', re.IGNORECASE)('A') == 'A'

    with pytest.raises(ValidationError):
        validate.Regexp(r'[0-9]+')('a')
    with pytest.raises(ValidationError):
        validate.Regexp(r'[a-z]+')('1')
    with pytest.raises(ValidationError):
        validate.Regexp(r'a')('A')

def test_regexp_compile():
    assert validate.Regexp(re.compile(r'a'))('a') == 'a'
    assert validate.Regexp(re.compile(r'\w'))('_') == '_'
    assert validate.Regexp(re.compile(r'\s'))(' ') == ' '
    assert validate.Regexp(re.compile(r'1'))('1') == '1'
    assert validate.Regexp(re.compile(r'[0-9]+'))('1') == '1'
    assert validate.Regexp(re.compile(r'a', re.IGNORECASE))('A') == 'A'
    assert validate.Regexp(re.compile(r'a', re.IGNORECASE), re.IGNORECASE)('A') == 'A'

    with pytest.raises(ValidationError):
        validate.Regexp(re.compile(r'[0-9]+'))('a')
    with pytest.raises(ValidationError):
        validate.Regexp(re.compile(r'[a-z]+'))('1')
    with pytest.raises(ValidationError):
        validate.Regexp(re.compile(r'a'))('A')
    with pytest.raises(ValidationError):
        validate.Regexp(re.compile(r'a'), re.IGNORECASE)('A')

def test_regexp_custom_message():
    rex = r'[0-9]+'
    v = validate.Regexp(rex, error='{input} does not match {regex}')
    with pytest.raises(ValidationError) as excinfo:
        v('a')
    assert 'a does not match [0-9]+' in str(excinfo)

def test_regexp_repr():
    assert (
        repr(validate.Regexp(regex='abc', flags=0, error=None)) ==
        '<Regexp(regex={0!r}, error={1!r})>'
        .format(re.compile('abc'), 'String does not match expected pattern.')
    )
    assert (
        repr(validate.Regexp(regex='abc', flags=re.IGNORECASE, error='foo')) ==
        '<Regexp(regex={0!r}, error={1!r})>'
        .format(re.compile('abc', re.IGNORECASE), 'foo')
    )


def test_predicate():
    class Dummy(object):
        def _true(self):
            return True

        def _false(self):
            return False

        def _list(self):
            return [1, 2, 3]

        def _empty(self):
            return []

        def _identity(self, arg):
            return arg

    d = Dummy()

    assert validate.Predicate('_true')(d) == d
    assert validate.Predicate('_list')(d) == d
    assert validate.Predicate('_identity', arg=True)(d) == d
    assert validate.Predicate('_identity', arg=1)(d) == d
    assert validate.Predicate('_identity', arg='abc')(d) == d

    with pytest.raises(ValidationError) as excinfo:
        validate.Predicate('_false')(d)
    assert 'Invalid input.' in str(excinfo)
    with pytest.raises(ValidationError):
        validate.Predicate('_empty')(d)
    with pytest.raises(ValidationError):
        validate.Predicate('_identity', arg=False)(d)
    with pytest.raises(ValidationError):
        validate.Predicate('_identity', arg=0)(d)
    with pytest.raises(ValidationError):
        validate.Predicate('_identity', arg='')(d)

def test_predicate_custom_message():
    class Dummy(object):
        def _false(self):
            return False

        def __str__(self):
            return 'Dummy'
    d = Dummy()
    with pytest.raises(ValidationError) as excinfo:
        validate.Predicate('_false', error='{input}.{method} is invalid!')(d)
    assert 'Dummy._false is invalid!' in str(excinfo)

def test_predicate_repr():
    assert (
        repr(validate.Predicate(method='foo', error=None)) ==
        '<Predicate(method={0!r}, kwargs={1!r}, error={2!r})>'
        .format('foo', {}, 'Invalid input.')
    )
    assert (
        repr(validate.Predicate(method='foo', error='bar', zoo=1)) ==
        '<Predicate(method={0!r}, kwargs={1!r}, error={2!r})>'
        .format('foo', {str('zoo') if PY2 else 'zoo': 1}, 'bar')
    )


def test_noneof():
    assert validate.NoneOf([1, 2, 3])(4) == 4
    assert validate.NoneOf('abc')('d') == 'd'
    assert validate.NoneOf('')([]) == []
    assert validate.NoneOf([])('') == ''
    assert validate.NoneOf([])([]) == []
    assert validate.NoneOf([1, 2, 3])(None) is None

    with pytest.raises(ValidationError) as excinfo:
        validate.NoneOf([1, 2, 3])(3)
    assert 'Invalid input.' in str(excinfo)
    with pytest.raises(ValidationError):
        validate.NoneOf('abc')('c')
    with pytest.raises(ValidationError):
        validate.NoneOf([1, 2, None])(None)
    with pytest.raises(ValidationError):
        validate.NoneOf('')('')

def test_noneof_custom_message():
    with pytest.raises(ValidationError) as excinfo:
        validate.NoneOf([1, 2], error='<not valid>')(1)
    assert '<not valid>' in str(excinfo)

    none_of = validate.NoneOf(
        [1, 2],
        error='{input} cannot be one of {values}'
    )
    with pytest.raises(ValidationError) as excinfo:
        none_of(1)
    assert '1 cannot be one of 1, 2' in str(excinfo)

def test_noneof_repr():
    assert (
        repr(validate.NoneOf(iterable=[1, 2, 3], error=None)) ==
        '<NoneOf(iterable=[1, 2, 3], error={0!r})>'
        .format('Invalid input.')
    )
    assert (
        repr(validate.NoneOf(iterable=[1, 2, 3], error='foo')) ==
        '<NoneOf(iterable=[1, 2, 3], error={0!r})>'
        .format('foo')
    )


def test_oneof():
    assert validate.OneOf([1, 2, 3])(2) == 2
    assert validate.OneOf('abc')('b') == 'b'
    assert validate.OneOf('')('') == ''
    assert validate.OneOf(dict(a=0, b=1))('a') == 'a'
    assert validate.OneOf((1, 2, None))(None) is None

    with pytest.raises(ValidationError) as excinfo:
        validate.OneOf([1, 2, 3])(4)
    assert 'Not a valid choice.' in str(excinfo)
    with pytest.raises(ValidationError):
        validate.OneOf('abc')('d')
    with pytest.raises(ValidationError):
        validate.OneOf((1, 2, 3))(None)
    with pytest.raises(ValidationError):
        validate.OneOf([])([])
    with pytest.raises(ValidationError):
        validate.OneOf(())(())
    with pytest.raises(ValidationError):
        validate.OneOf(dict(a=0, b=1))(0)
    with pytest.raises(ValidationError):
        validate.OneOf('123')(1)

def test_oneof_options():
    oneof = validate.OneOf([1, 2, 3], ['one', 'two', 'three'])
    expected = [('1', 'one'), ('2', 'two'), ('3', 'three')]
    assert list(oneof.options()) == expected

    oneof = validate.OneOf([1, 2, 3], ['one', 'two'])
    expected = [('1', 'one'), ('2', 'two'), ('3', '')]
    assert list(oneof.options()) == expected

    oneof = validate.OneOf([1, 2], ['one', 'two', 'three'])
    expected = [('1', 'one'), ('2', 'two'), ('', 'three')]
    assert list(oneof.options()) == expected

    oneof = validate.OneOf([1, 2])
    expected = [('1', ''), ('2', '')]
    assert list(oneof.options()) == expected

def test_oneof_text():
    oneof = validate.OneOf([1, 2, 3], ['one', 'two', 'three'])
    assert oneof.choices_text == '1, 2, 3'
    assert oneof.labels_text == 'one, two, three'

    oneof = validate.OneOf([1], ['one'])
    assert oneof.choices_text == '1'
    assert oneof.labels_text == 'one'

    oneof = validate.OneOf(dict(a=0, b=1))
    assert ', '.join(sorted(oneof.choices_text.split(', '))) == 'a, b'
    assert oneof.labels_text == ''

def test_oneof_custom_message():
    oneof = validate.OneOf([1, 2, 3], error='{input} is not one of {choices}')
    expected = '4 is not one of 1, 2, 3'
    with pytest.raises(ValidationError):
        oneof(4)
    assert expected in str(expected)

    oneof = validate.OneOf([1, 2, 3],
        ['one', 'two', 'three'],
        error='{input} is not one of {labels}'
    )
    expected = '4 is not one of one, two, three'
    with pytest.raises(ValidationError):
        oneof(4)
    assert expected in str(expected)

def test_oneof_repr():
    assert (
        repr(validate.OneOf(choices=[1, 2, 3], labels=None, error=None)) ==
        '<OneOf(choices=[1, 2, 3], labels=[], error={0!r})>'
        .format('Not a valid choice.')
    )
    assert (
        repr(validate.OneOf(choices=[1, 2, 3], labels=['a', 'b', 'c'], error='foo')) ==
        '<OneOf(choices=[1, 2, 3], labels={0!r}, error={1!r})>'
        .format(['a', 'b', 'c'], 'foo')
    )


def test_containsonly_in_list():
    assert validate.ContainsOnly([])([]) == []
    assert validate.ContainsOnly([1, 2, 3])([1]) == [1]
    assert validate.ContainsOnly([1, 1, 2])([1, 1]) == [1, 1]
    assert validate.ContainsOnly([1, 2, 3])([1, 2]) == [1, 2]
    assert validate.ContainsOnly([1, 2, 3])([2, 1]) == [2, 1]
    assert validate.ContainsOnly([1, 2, 3])([1, 2, 3]) == [1, 2, 3]
    assert validate.ContainsOnly([1, 2, 3])([3, 1, 2]) == [3, 1, 2]
    assert validate.ContainsOnly([1, 2, 3])([2, 3, 1]) == [2, 3, 1]
    assert validate.ContainsOnly([1, 2, 3])([1, 2, 3, 1]) == [1, 2, 3, 1]
    assert validate.ContainsOnly([1, 2, 3])([]) == []

    with pytest.raises(ValidationError):
        validate.ContainsOnly([1, 2, 3])([4])
    with pytest.raises(ValidationError):
        validate.ContainsOnly([])([1])

def test_contains_only_unhashable_types():
    assert validate.ContainsOnly([[1], [2], [3]])([[1]]) == [[1]]
    assert validate.ContainsOnly([[1], [1], [2]])([[1], [1]]) == [[1], [1]]
    assert validate.ContainsOnly([[1], [2], [3]])([[1], [2]]) == [[1], [2]]
    assert validate.ContainsOnly([[1], [2], [3]])([[2], [1]]) == [[2], [1]]
    assert validate.ContainsOnly([[1], [2], [3]])([[1], [2], [3]]) == [[1], [2], [3]]
    assert validate.ContainsOnly([[1], [2], [3]])([[3], [1], [2]]) == [[3], [1], [2]]
    assert validate.ContainsOnly([[1], [2], [3]])([[2], [3], [1]]) == [[2], [3], [1]]
    assert validate.ContainsOnly([[1], [2], [3]])([]) == []

    with pytest.raises(ValidationError):
        validate.ContainsOnly([[1], [2], [3]])([[4]])
    with pytest.raises(ValidationError):
        validate.ContainsOnly([])([1])

def test_containsonly_in_tuple():
    assert validate.ContainsOnly(())(()) == ()
    assert validate.ContainsOnly((1, 2, 3))((1,)) == (1,)
    assert validate.ContainsOnly((1, 1, 2))((1, 1)) == (1, 1)
    assert validate.ContainsOnly((1, 2, 3))((1, 2)) == (1, 2)
    assert validate.ContainsOnly((1, 2, 3))((2, 1)) == (2, 1)
    assert validate.ContainsOnly((1, 2, 3))((1, 2, 3)) == (1, 2, 3)
    assert validate.ContainsOnly((1, 2, 3))((3, 1, 2)) == (3, 1, 2)
    assert validate.ContainsOnly((1, 2, 3))((2, 3, 1)) == (2, 3, 1)
    assert validate.ContainsOnly((1, 2, 3))(()) == tuple()
    with pytest.raises(ValidationError):
        validate.ContainsOnly((1, 2, 3))((4,))
    with pytest.raises(ValidationError):
        validate.ContainsOnly(())((1,))


def test_contains_only_in_string():
    assert validate.ContainsOnly('')('') == ''
    assert validate.ContainsOnly('abc')('a') == 'a'
    assert validate.ContainsOnly('aab')('aa') == 'aa'
    assert validate.ContainsOnly('abc')('ab') == 'ab'
    assert validate.ContainsOnly('abc')('ba') == 'ba'
    assert validate.ContainsOnly('abc')('abc') == 'abc'
    assert validate.ContainsOnly('abc')('cab') == 'cab'
    assert validate.ContainsOnly('abc')('bca') == 'bca'
    assert validate.ContainsOnly('abc')('') == ''

    with pytest.raises(ValidationError):
        validate.ContainsOnly('abc')('d')
    with pytest.raises(ValidationError):
        validate.ContainsOnly('')('a')


def test_containsonly_custom_message():
    containsonly = validate.ContainsOnly(
        [1, 2, 3],
        error='{input} is not one of {choices}'
    )
    expected = '4, 5 is not one of 1, 2, 3'
    with pytest.raises(ValidationError):
        containsonly([4, 5])
    assert expected in str(expected)

    containsonly = validate.ContainsOnly([1, 2, 3],
        ['one', 'two', 'three'],
        error='{input} is not one of {labels}'
    )
    expected = '4, 5 is not one of one, two, three'
    with pytest.raises(ValidationError):
        containsonly([4, 5])
    assert expected in str(expected)

def test_containsonly_repr():
    assert (
        repr(validate.ContainsOnly(choices=[1, 2, 3], labels=None, error=None)) ==
        '<ContainsOnly(choices=[1, 2, 3], labels=[], error={0!r})>'
        .format('One or more of the choices you made was not acceptable.')
    )
    assert (
        repr(validate.ContainsOnly(choices=[1, 2, 3], labels=['a', 'b', 'c'], error='foo')) ==
        '<ContainsOnly(choices=[1, 2, 3], labels={0!r}, error={1!r})>'
        .format(['a', 'b', 'c'], 'foo')
    )



# -*- coding: utf-8 -*-
from collections import OrderedDict
import datetime as dt

import pytest

from marshmallow import fields, Schema

from tests.base import User


class TestUnordered:

    class UnorderedSchema(Schema):
        name = fields.Str()
        email = fields.Str()

        class Meta:
            ordered = False

    def test_unordered_dump_returns_dict(self):
        schema = self.UnorderedSchema()
        u = User('steve', email='steve@steve.steve')
        result = schema.dump(u)
        assert not isinstance(result, OrderedDict)
        assert type(result) is dict

    def test_unordered_load_returns_dict(self):
        schema = self.UnorderedSchema()
        result = schema.load({'name': 'steve', 'email': 'steve@steve.steve'})
        assert not isinstance(result, OrderedDict)
        assert type(result) is dict


class KeepOrder(Schema):
    class Meta:
        ordered = True

    name = fields.String(allow_none=True)
    email = fields.Email(allow_none=True)
    age = fields.Integer()
    created = fields.DateTime()
    id = fields.Integer(allow_none=True)
    homepage = fields.Url()
    birthdate = fields.Date()

class OrderedMetaSchema(Schema):
    id = fields.Int(allow_none=True)
    email = fields.Email(allow_none=True)

    class Meta:
        fields = ('name', 'email', 'age', 'created',
                    'id', 'homepage', 'birthdate')
        ordered = True

class OrderedNestedOnly(Schema):
    class Meta:
        ordered = True

    user = fields.Nested(KeepOrder)

class TestFieldOrdering:

    @pytest.mark.parametrize('with_meta', (False, True))
    def test_ordered_option_is_inherited(self, user, with_meta):
        class ParentUnordered(Schema):
            class Meta:
                ordered = False

        # KeepOrder is before ParentUnordered in MRO,
        # so ChildOrderedSchema will be ordered
        class ChildOrderedSchema(KeepOrder, ParentUnordered):
            if with_meta:
                class Meta:
                    pass

        schema = ChildOrderedSchema()
        assert schema.opts.ordered is True
        assert schema.dict_class == OrderedDict

        data = schema.dump(user)
        keys = list(data)
        assert keys == ['name', 'email', 'age', 'created', 'id', 'homepage', 'birthdate']

        # KeepOrder is before ParentUnordered in MRO,
        # so ChildOrderedSchema will be ordered
        class ChildUnorderedSchema(ParentUnordered, KeepOrder):
            class Meta:
                pass

        schema = ChildUnorderedSchema()
        assert schema.opts.ordered is False

    def test_ordering_is_off_by_default(self):
        class DummySchema(Schema):
            pass

        schema = DummySchema()
        assert schema.ordered is False

    def test_declared_field_order_is_maintained_on_dump(self, user):
        ser = KeepOrder()
        data = ser.dump(user)
        keys = list(data)
        assert keys == ['name', 'email', 'age', 'created', 'id', 'homepage', 'birthdate']

    def test_declared_field_order_is_maintained_on_load(self, serialized_user):
        schema = KeepOrder()
        data = schema.load(serialized_user)
        keys = list(data)
        assert keys == ['name', 'email', 'age', 'created', 'id', 'homepage', 'birthdate']

    def test_nested_field_order_with_only_arg_is_maintained_on_dump(self, user):
        schema = OrderedNestedOnly()
        data = schema.dump({'user': user})
        user_data = data['user']
        keys = list(user_data)
        assert keys == ['name', 'email', 'age', 'created', 'id', 'homepage', 'birthdate']

    def test_nested_field_order_with_only_arg_is_maintained_on_load(self):
        schema = OrderedNestedOnly()
        data = schema.load({'user': {
            'name': 'Foo',
            'email': 'Foo@bar.com',
            'age': 42,
            'created': dt.datetime.now().isoformat(),
            'id': 123,
            'homepage': 'http://foo.com',
            'birthdate': dt.datetime.now().isoformat(),
        }})
        user_data = data['user']
        keys = list(user_data)
        assert keys == ['name', 'email', 'age', 'created', 'id', 'homepage', 'birthdate']

    def test_nested_field_order_with_exclude_arg_is_maintained(self, user):
        class HasNestedExclude(Schema):
            class Meta:
                ordered = True

            user = fields.Nested(KeepOrder, exclude=('birthdate', ))

        ser = HasNestedExclude()
        data = ser.dump({'user': user})
        user_data = data['user']
        keys = list(user_data)
        assert keys == ['name', 'email', 'age', 'created', 'id', 'homepage']

    def test_meta_fields_order_is_maintained_on_dump(self, user):
        ser = OrderedMetaSchema()
        data = ser.dump(user)
        keys = list(data)
        assert keys == ['name', 'email', 'age', 'created', 'id', 'homepage', 'birthdate']

    def test_meta_fields_order_is_maintained_on_load(self, serialized_user):
        schema = OrderedMetaSchema()
        data = schema.load(serialized_user)
        keys = list(data)
        assert keys == ['name', 'email', 'age', 'created', 'id', 'homepage', 'birthdate']

class TestIncludeOption:

    class AddFieldsSchema(Schema):
        name = fields.Str()

        class Meta:
            include = {
                'from': fields.Str()
            }

    def test_fields_are_added(self):
        s = self.AddFieldsSchema()
        in_data = {'name': 'Steve', 'from': 'Oskosh'}
        result = s.load({'name': 'Steve', 'from': 'Oskosh'})
        assert result == in_data

    def test_ordered_included(self):
        class AddFieldsOrdered(Schema):
            name = fields.Str()
            email = fields.Str()

            class Meta:
                include = OrderedDict([
                    ('from', fields.Str()),
                    ('in', fields.Str()),
                    ('@at', fields.Str())
                ])
                ordered = True

        s = AddFieldsOrdered()
        in_data = {'name': 'Steve', 'from': 'Oskosh', 'email': 'steve@steve.steve',
                    'in': 'VA', '@at': 'Charlottesville'}
        # declared fields, then "included" fields
        expected_fields = ['name', 'email', 'from', 'in', '@at']
        assert list(AddFieldsOrdered._declared_fields.keys()) == expected_fields

        result = s.load(in_data)
        assert list(result.keys()) == expected_fields

    def test_added_fields_are_inherited(self):

        class AddFieldsChild(self.AddFieldsSchema):
            email = fields.Str()

        s = AddFieldsChild()
        assert 'email' in s._declared_fields.keys()
        assert 'from' in s._declared_fields.keys()
        assert isinstance(s._declared_fields['from'], fields.Str)
